/**
 * A helper that checks if the given argument is of a certain type.
 */
export function $is(type: BigIntConstructor, arg: unknown): arg is bigint;
export function $is(type: BooleanConstructor, arg: unknown): arg is boolean;
export function $is(type: NumberConstructor, arg: unknown): arg is number;
export function $is(type: StringConstructor, arg: unknown): arg is string;
export function $is(type: SymbolConstructor, arg: unknown): arg is symbol;
export function $is(type: undefined, arg: unknown): arg is null | undefined;
export function $is<T extends abstract new (...args: any[]) => any>(type: T, arg: unknown): arg is InstanceType<T>;

/**
 * Replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.
 */
export function $in(key: any, ref: any): boolean;

/**
 * A helper that checks equality of the given arguments.
 */
export function $eq(x: any, y: any): boolean;

export type $HArgs = [
  keyof HTMLElementTagNameMap | Element | ShadowRoot,
  Record<string, any>,
  ...($HArgs | Element | ShadowRoot | Text | string | number)[]
];

/**
 * A lightweight ExtJS-like helper for creating and modifying `HTMLElement`s.
 *
 * - `e`: the modified `HTMLElement` or tag of the element being created.
 * - `attributesAndProperties`: map of attributes and properties. Keys starting with `$` are treated as *properties* to be set on the element being created or modified (without the leading `$`). All other keys are treated as *attributes* to be set by `setAttribute`. Attributes equal to `false` cause the attribute to be removed by `removeAttribute`.
 * - `items`: items to be appended to the element being created or modified.
 */
export function $h<T extends keyof HTMLElementTagNameMap>(e: T, keys?: $HArgs[1], ...items: $HArgs[2][]): HTMLElementTagNameMap[T];
export function $h<E extends Element>(e: E, keys?: $HArgs[1], ...items: $HArgs[2][]): E;
export function $h<S extends ShadowRoot>(e: S, keys?: $HArgs[1], ...items: $HArgs[2][]): S;
export function $h(e: $HArgs[0], keys?: $HArgs[1], ...items: $HArgs[2][]): Element | ShadowRoot;

export type $SArgs = [
  keyof SVGElementTagNameMap | SVGElement,
  Record<string, any>,
  ...($SArgs | SVGElement | Text | string | number)[]
];

/**
 * A lightweight ExtJS-like helper for creating and modifying `SVGElement`s (conf. `$h`).
 *
 * - `e`: the modified `SVGElement` or tag of the element being created.
 * - `attributesAndProperties`: map of attributes and properties. Keys starting with `$` are treated as *properties* to be set on the element being created or modified (without the leading `$`). All other keys are treated as *attributes* to be set by `setAttribute` or `setAttributeNS`. Attributes equal to `false` cause the attribute to be removed by `removeAttribute` or `removeAttributeNS`.
 * - `items`: items to be appended to the element being created or modified.
 */
export function $s<T extends keyof SVGElementTagNameMap>(e: T, keys?: $SArgs[1], ...items: $SArgs[2][]): SVGElementTagNameMap[T];
export function $s<E extends SVGElement>(e: E, keys?: $SArgs[1], ...items: $SArgs[2][]): E;
export function $s(e: $SArgs[0], keys?: $SArgs[1], ...items: $SArgs[2][]): SVGElement;

export type $CNode = {
  [attributeOrSelector: string]: string | number | $CNode;
};

/**
 * Simple CSS-in-JS helper.
 * An iterative implementation that does not cause a stack overflow exception.
 *
 * The `root` parameter provides a hierarchical description of CSS rules.
 *
 * - Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
 * - All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
 * - In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
 * - Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
 * - Top-level keys that begin with `@` are not concatenated with sub-object keys.
 */
export function $c(root: Record<string, $CNode>, splitter?: string): string;

/**
 * A generic helper for syntax highlighting.
 */
export function $highlight(pattern: RegExp, highlighters: ((chunk: string, index: number) => $HArgs[2])[], innerText: string): $HArgs[2][];

/**
 * Language translations helper.
 */
export function $locale(
  locales: Record<string, Record<string, string | Record<string, string>>>,
  defaultLanguage: string,
  languages?: readonly string[]
): (text?: string | undefined, version?: string | undefined) => string;
